use std::sync::Arc;

use datatype::DataType;
use serde::ser::{SerializeMap, Serializer};
use serde::{Deserialize, Serialize};
use validation::Validation;

mod datatype;
mod validation;

#[derive(Clone, Debug)]
pub struct Parameter {
    name: Arc<str>,

    data_type: DataType,

    #[allow(dead_code)]
    is_array: bool,

    aliases: Vec<Arc<str>>,
    description: Option<Arc<str>>,
    help_message: Option<Arc<str>>,
    parameter_sets: Vec<ParameterSet>,
    validations: Vec<Validation>,
}

impl Serialize for Parameter {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut state = serializer.serialize_map(None)?;

        state.serialize_entry("name", &self.name)?;

        let mut type_str = format!("{}", self.data_type);
        if self.is_array {
            type_str.push_str("[]");
        }
        state.serialize_entry("type", &type_str)?;

        state.serialize_entry("aliases", &self.aliases)?;

        if let Some(desc) = &self.description {
            state.serialize_entry("description", desc)?;
        }

        if let Some(help) = &self.help_message {
            state.serialize_entry("helpMessage", help)?;
        }

        state.serialize_entry("parameterSets", &self.parameter_sets)?;
        state.serialize_entry("validations", &self.validations)?;

        state.end()
    }
}

impl<'de> Deserialize<'de> for Parameter {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        #[derive(Deserialize)]
        #[serde(rename_all = "camelCase")]
        struct RawParameter {
            name: Arc<str>,
            #[serde(rename = "type")]
            data_type: String,
            aliases: Vec<Arc<str>>,
            description: Option<Arc<str>>,
            help_message: Option<Arc<str>>,
            parameter_sets: Vec<ParameterSet>,
            validations: Vec<Validation>,
        }

        let raw = RawParameter::deserialize(deserializer)?;

        let is_array = raw.data_type.ends_with("[]");
        let type_str = raw.data_type.trim_end_matches("[]");

        let data_type: DataType =
            serde_json::from_str(&format!("\"{}\"", type_str)).map_err(serde::de::Error::custom)?;

        Ok(Parameter {
            name: raw.name,
            data_type,
            is_array,
            aliases: raw.aliases,
            description: raw.description,
            help_message: raw.help_message,
            parameter_sets: raw.parameter_sets,
            validations: raw.validations,
        })
    }
}

#[derive(Deserialize, Serialize, Clone, PartialEq, Eq, Default, Debug)]
pub struct ParameterSet {
    name: Arc<str>,
    mandatory: bool,
    position: Option<u8>,
    value_from_pipeline: Option<bool>,
    value_from_pipeline_by_property_name: Option<bool>,
}

#[cfg(test)]
mod tests {
    #![allow(clippy::unwrap_used, clippy::expect_used)]
    use super::*;
    use serde_json::{Value, json};

    fn make_base_parameter(is_array: bool, data_type: DataType) -> Parameter {
        Parameter {
            name: Arc::from("testParam"),
            data_type,
            is_array,
            aliases: vec![Arc::from("alias1")],
            description: Some(Arc::from("desc")),
            help_message: Some(Arc::from("help")),
            parameter_sets: vec![],
            validations: vec![],
        }
    }

    #[test]
    fn deserialize_with_array_type() {
        let json_data = json!({
            "name": "foo",
            "type": "string[]",
            "aliases": ["aliasA"],
            "description": "some description",
            "helpMessage": "some help",
            "parameterSets": [],
            "validations": []
        });

        let param: Parameter = serde_json::from_value(json_data).unwrap();
        assert!(param.is_array);
        assert!(matches!(param.data_type, DataType::String));
    }

    #[test]
    fn deserialize_without_array_type() {
        let json_data = json!({
            "name": "bar",
            "type": "boolean",
            "aliases": [],
            "description": null,
            "helpMessage": null,
            "parameterSets": [],
            "validations": []
        });

        let param: Parameter = serde_json::from_value(json_data).unwrap();
        assert!(!param.is_array);
        assert!(matches!(param.data_type, DataType::Boolean));
    }

    #[test]
    fn serialize_with_array_type() {
        let param = make_base_parameter(true, DataType::Int32);
        let json_str = serde_json::to_string(&param).unwrap();
        let value: Value = serde_json::from_str(&json_str).unwrap();

        assert_eq!(value["type"], "Int32[]");
        assert_eq!(value["name"], "testParam");
        assert!(value.get("is_array").is_none()); // is_array must not be serialized
    }

    #[test]
    fn serialize_without_array_type() {
        let param = make_base_parameter(false, DataType::Int64);
        let json_str = serde_json::to_string(&param).unwrap();
        let value: Value = serde_json::from_str(&json_str).unwrap();

        assert_eq!(value["type"], "Int64");
        assert!(value.get("is_array").is_none()); // is_array must not be serialized
    }
}
